#!/usr/bin/env python

"""
Material:
    * 0.84mm fiber: 400m (200x2m)
    * 1.5mm fiber : 30m
    * 2mm fiber   : 30m
    * 2.5mm fiber : 20m
    * 3mm fiber   : 20m

One led driver with VA 15V can power 3 5V leds.
These led drivers need to be powered through ULN2803As with PWM as inputs.
Each arduino has 6 PWM pins.
If we use 8 minis we get 6 twinkle fields for a maximum distance of 80cm.
Which should give us a maximum fiber length of 100cm.
Assuming an equal spread this means an average length of 75cm.

Stars:
    * 0.84mm: 530
    * 1.5mm : 40
    * 2mm   : 40
    * 2.5mm : 25
    * 3mm   : 25
            + ===
              660
"""

import sys
import math
import random
import os.path


# diameter as part of an area in the field (for drawing purposes)
SIZES = [0.084, 0.15, 0.2, 0.25, 0.3]

STARS = [530, 40, 40, 25, 25]

# field size
X = 400
Y = 244

# the number of pixels per area in the field
FACTOR = 20

# the base chance to have a star appear in a non-pulled area
BASE = 100

# gravity pulls ((x, y), a, b)
# with (x, y): the coordinates of the gravity center
#      a: the power (the greater the power, the more stars in the pull)
#      b: the density (the distance at which the pull loses its effect)
# the density is the number of spaces away the pull will have any effect
#
# with a density of 50 and a power of 10 you'll have 11 times more chance of
# a star at the center of the pull, 6 times more chance at 25 distance from the
# pull, and 2 times more chance at 45 distance
# with a density of 100 and a power of 5 you'll have 6 times more chance of
# a star at the center of the pull, 4.75 times more chance at 25, 3.5 times
# more chance at 50 and 1.5 times more chance at 90
GRAVITY = [
    ((300, 244), 10, 30),
    ((350, 240), 10, 60),
    ((450, 230), 10, 125),
    ((150, 40), 10, 20),
    ((125, 45), 10, 35),
    ((75, 60), 10, 55),
    ((50, 100), 10, 60),
    ((40, 140), 10, 55),
    ((55, 200), 10, 80),
    ((150, 200), 10, 60),
    ((155, 165), 10, 25),
    ((150, 135), 10, 20),
    ((140, 125), 10, 10),
]


# the way the power of the gravity pull dissipates over distance
def modulate_power(power, density, distance):
    return linear(power, density, distance)


# linear dissipation (distance = half density: modulated power = half power)
def linear(power, density, distance):
    return power - power * (distance/density)


def printmatrix(matrix):
    for i in range(len(matrix)):
        line = matrix[i]
        if i == 0:
            print(' '.join(['  '] + [str(x).zfill(2)
                                     for x in range(len(line))]))
        print(' '.join([str(i).zfill(2)] + [str(int(x/100)).zfill(2)
                                            for x in line]))


def printstars(stars):
    for i in range(len(stars)):
        line = stars[i]
        if i == 0:
            print(' '.join(['  '] + [str(x).zfill(2)
                                     for x in range(len(line))]))
        print(' '.join([str(i).zfill(2)]+line))


def buildmatrix():
    # build the matrix of chances
    matrix = [[BASE for x in range(X)] for y in range(Y)]
    # apply the gravity pulls
    for coords, power, density in GRAVITY:
        x, y = coords
        for a in range(max(0, x - density), min(X, x + density + 1)):
            for b in range(max(0, y - density), min(Y, y + density + 1)):
                distance = math.sqrt(abs(x - a) ** 2 + abs(y - b) ** 2)
                if distance < density:
                    modulated_power = modulate_power(power, density, distance)
                    matrix[b][a] += int(modulated_power * BASE)
    return matrix


def buildstars(matrix):
    # builds a star map using the given matrix of chances
    stars = [['  ' for x in range(X)] for y in range(Y)]
    coords_done = []
    chance_array = []
    cur = 0
    for x in range(X):
        for y in range(Y):
            cur += matrix[y][x]
            chance_array += [(cur, (x, y))]
    for n_size in range(len(STARS)):
        size = SIZES[n_size]
        for n in range(STARS[n_size]):
            while True:
                filled = False
                r = random.randint(0, cur-1)
                for chance, coords in chance_array:
                    if r < chance:
                        if not (size, coords) in coords_done:
                            x, y = coords
                            stars[y][x] = ' {}'.format(n_size)
                            coords_done += [(size, coords)]
                            filled = True
                        break
                if filled:
                    break
    coords_done.sort()
    return stars, coords_done


def create_image(coords):
    from PIL import Image, ImageDraw

    image = Image.new('1', (X * FACTOR + FACTOR, Y * FACTOR + FACTOR))
    draw = ImageDraw.Draw(image)
    for size, coord in coords:
        dia = int(round(size * FACTOR * 3, 0))
        x, y = [a * FACTOR + FACTOR / 2 for a in coord]
        x1, y1 = x - dia/2, y - dia/2
        x2, y2 = x + dia/2, y + dia/2
        draw.ellipse((x1, y1, x2, y2), fill=255, outline=255)
    del draw
    image.save('starmap.png')


if __name__ == '__main__':
    if (not len(sys.argv) > 1 or
            '-h' in sys.argv or
            '--help' in sys.argv or
            '-?' in sys.argv):
        print('USAGE: {} (matrix) (stars) (coords) (panels) (image)'.format(
                os.path.split(sys.argv[0])[1]))
        sys.exit(0)
    matrix = buildmatrix()
    stars, coords = buildstars(matrix)
    # allstars = []
    # for line in stars: allstars += line
    # for size in range(len(STARS)):
    #     print(size, allstars.count(' {}'.format(size)))
    for el in sys.argv[1:]:
        if el.lower() == 'matrix':
            printmatrix(matrix)
        elif el.lower() == 'stars':
            printstars(stars)
        elif el.lower() == 'coords':
            for size, coord in coords:
                print(size, '@', coord)
        elif el.lower() == 'panels':
            # split coords up in four panels
            panels = [[], [], [], []]
            for size, coord in coords:
                x, y = coord
                if x < X / 2 and y < Y / 2:
                    panels[0] += [(size, coord)]
                elif x >= X / 2 and y < Y / 2:
                    panels[1] += [(size, coord)]
                elif x < X / 2 and y >= Y / 2:
                    panels[2] += [(size, coord)]
                else:
                    panels[3] += [(size, coord)]
            for i in range(len(panels)):
                print('PANEL {}:'.format(i + 1))
                for size, coord in panels[i]:
                    print(size, '@', coord)
                if not i + 1 == len(panels):
                    print('')
        elif el.lower() == 'image':
            create_image(coords)
